import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import ergastirio.panels
import ergastirio.widgets
import ergastirio.utils
import ergastirio.styles
from importlib.metadata import version  

class experiment_gui():
    def __init__(self, exp, parent):
        self.exp = exp
        self.app = self.exp.app
        self.parent = parent 
        self.logger = self.exp.logger

        self.set_up_experiment_gui()
        self.exp.fire_all_signals() #ask the experiment to "fire" all of its signals, in order to communicate to different parts of the GUI the current values of several parameters

    def set_up_experiment_gui(self):
        self.container_connectramps = self.parent.advanced_panels['connect_ramps']['container']
        self.container_tabledata = self.parent.basic_panels['tabledata']['container']
        self.container_plots = self.parent.basic_panels['plots']['scrollarea']
        self.container_instruments = self.parent.basic_panels['instruments']['container']

        self.create_gui_table_data()
        self.create_gui_plots()
    
        if not self.create_gui_instruments(): #Create GUIs for all loaded instruments 
            return False

        self.create_gui_connectramps()

        self.link_data_to_gui()

        #if 'DefaultView' in exp.config.keys():
        #    QtCore.QTimer.singleShot(1, lambda : exp.mainwindow.set_view(exp.config['DefaultView']))

        return True

    def create_gui_connectramps(self):
        self.connectramps_panel_box = Qt.QGroupBox()
        layout = Qt.QVBoxLayout()
        self.connectramps_panel = ergastirio.panels.ramp_connection( parent = self.connectramps_panel_box ,
                                                                    experiment = self.exp)
        layout.addWidget(self.connectramps_panel_box)
        self.container_connectramps.setLayout(layout)
        self.connectramps_panel.create_gui()

    def create_gui_instruments(self):
        '''
        Create GUIs for all loaded instruments, and for the acquisition control panel

        container
            QWidget which will contain the GUIs
        '''

        #if ('Alignment_Instruments_Window' in self.exp.config.keys())  and self.exp.config['Alignment_Instruments_Window']=='H':
        #    box = Qt.QHBoxLayout()
        #else: 
        box = Qt.QVBoxLayout()

        self.logger.info(f"Creating GUI for all loaded instruments...")
    
        for i,instrument in enumerate(self.exp.instruments):
            self.exp.instruments[i]['frame'] = Qt.QGroupBox() # Qt.QWidget()
            self.exp.instruments[i]['frame'].setObjectName(self.exp.instruments[i]['name']+'_frame')
            string = "QGroupBox#"+ self.exp.instruments[i]['frame'].objectName()
            self.exp.instruments[i]['frame'].setStyleSheet(string + ergastirio.styles.instrument_box + string + ergastirio.styles.instrument_boxtitle) #This line changes the style of ONLY this QWdiget

            self.exp.instruments[i]['frame'].setTitle(f"{self.exp.instruments[i]['name']} ({self.exp.instruments[i]['type']} v{version(self.exp.instruments[i]['type'])})")
        
            self.exp.instruments[i]['gui'] = self.exp.instruments[i]['device_module'].gui(interface = self.exp.instruments[i]['interface'],parent=self.exp.instruments[i]['frame'])

            self.exp.instruments[i]['frame'].resize(self.exp.instruments[i]['frame'].sizeHint())
            box.addWidget(self.exp.instruments[i]['frame'] )
        box.addStretch(1)

        self.container_instruments.setLayout(box) #This line makes sure that all widgest defined so far are assigned to the widget defined in container
        self.container_instruments.resize(self.container_instruments.sizeHint())
        return True

    def create_gui_plots(self):
        '''
        Create GUIs for all plots.
        Note: we keep two separated list, self.plots and self.exp.plots
        self.plots[plotindex] contains all the GUI objects for the plot identified by index = plotindex
        self.exp.plots[plotindex] contains all the non-GUI objects for the plot identified by index = plotindex

        '''
        self.logger.info(f"Creating GUI for all plots...")

        mdi = Qt.QMdiArea() #Create an mdi area which will contain the plots
        self.plots =[]
        for plotindex, plot in enumerate(self.exp.plots):
            self.plots.append(dict())
            #For each plot we define a QMdiSubWindow, which will contain a QScrollArea, which will contain a QWidget
            # The QWidget will be used as a container for the actual plot
            self.plots[plotindex]['subwindow'] = Qt.QMdiSubWindow() 
            self.plots[plotindex]['subwindow'].setWindowFlags(QtCore.Qt.FramelessWindowHint)
            self.plots[plotindex]['subwindow'].setWindowTitle(self.exp.plots[plotindex]['name'])
            self.plots[plotindex]['scrollarea'] = Qt.QScrollArea()
            self.plots[plotindex]['subwindow'].setWidget(self.plots[plotindex]['scrollarea'])
            self.plots[plotindex]['widget'] = Qt.QWidget()
            self.plots[plotindex]['scrollarea'].setWidget(self.plots[plotindex]['widget'])
            self.plots[plotindex]['scrollarea'].setWidgetResizable(True)
            mdi.addSubWindow(self.plots[plotindex]['subwindow'])
            self.plots[plotindex]['plotobject'] = ergastirio.widgets.PlotObject(parent = self.plots[plotindex]['widget'],
                                                                               data_headers = self.exp.data_headers,
                                                                               plot_config = self.exp.plots[plotindex])
            self.plots[plotindex]['widget'].show()
    
        self.container_plots.mdi = mdi
        self.container_plots.setWidget(mdi)
        self.container_plots.mdi.tileSubWindows()
        return True

    def create_gui_table_data(self):
        '''
        Create GUIs for data acquisition, table, data management

        container
            QWidget which will contain the GUIs
        '''  

        layout = Qt.QVBoxLayout()
        layout_row1 = Qt.QHBoxLayout()
        layout_row2 = Qt.QHBoxLayout()

        ## CREATE ACQUISITION CONTROL PANEL
        self.logger.info(f"Creating GUI for acquisition control panel...")
        self.control_panel_container = Qt.QGroupBox() 
        self.control_panel_container.setObjectName('acquisition_panel')
        string = 'QGroupBox#acquisition_panel'
        self.control_panel_container.setStyleSheet(string + ergastirio.styles.acquisitionpanel_box + '\nQGroupBox' + ergastirio.styles.acquisitionpanel_boxtitle ) #This line changes the style of ONLY this QWdiget
        self.control_panel_container.setTitle(f"Acquisition Panel")
        self.control_panel = ergastirio.panels.acquisition_control( parent = self.control_panel_container,
                                                                    experiment = self.exp)
        self.control_panel.create_gui()
        ####

        ## CREATE DATA MANAGEMENT PANEL
        self.data_management_panel_container = Qt.QGroupBox()
        self.data_management_panel = ergastirio.panels.data_management( parent = self.data_management_panel_container,
                                                                        experiment = self.exp)
        self.data_management_panel.create_gui()
        ####

        ### CREATE TABLE TO SHOW DATA
        self.logger.info(f"Creating GUI for table...")
        self.tabledata_container = Qt.QGroupBox() 
        self.tabledata = ergastirio.widgets.Table(parent = self.tabledata_container)
        self.tabledata.data_headers = self.exp.data_headers
        self.tabledata.data = self.exp.data
        self.tabledata.create_gui()
        ###
    
        layout_row1.addWidget(self.control_panel_container)
        layout_row1.addWidget(self.data_management_panel_container)
        layout_row1.addStretch(1)
        layout_row2.addWidget(self.tabledata_container,stretch=3)
        layout.addLayout(layout_row1)
        layout.addLayout(layout_row2)
        self.container_tabledata.setLayout(layout)
 
        return True

    def link_data_to_gui(self):
        '''
        The data collected in the experiment is stored in self.exp.data, which is an instance of the EnhancedList class.
        This is a 'dynamic' list, wich allows 'linked objects'. Whenever any data stored in an EnhancedList object is changed, any linked
        object is notified, and a copy of the currently stored date is sent to the object.
        This is used as an elegant way to keep the table data and plots always syncronized with the stored data.
        To link an object, we use the method add_syncronized_objects defined in the EnhancedList class

        data.add_syncronized_objects([ InstanceOfTargetObject,  TargetClassProperty])

        Every time that the content of data is changed (e.g. a row is added to acquired data), the object data is also copied into
        the propery InstanceOfTargetObject.TargetClassProperty
        '''
        #self.exp.data.add_syncronized_objects([ self.tabledata,  ergastirio.widgets.Table.data])
        self.tabledata.add_signals(
                                    signal_data_updated = self.exp.sig_data_updated,
                                    signal_added_new_rows = self.exp.sig_data_updated_added_new_rows,
                                    signal_deleted_all_data = self.exp.sig_data_updated_deleted_all_data,
                                    signal_deleted_last_rows = self.exp.sig_data_updated_deleted_last_rows)
        for plot in self.plots:
            self.exp.data.add_syncronized_objects([  plot['plotobject'],  ergastirio.widgets.PlotObject.data])
            self.exp.data_std.add_syncronized_objects([  plot['plotobject'],  ergastirio.widgets.PlotObject.data_error])




import PyQt5.QtWidgets as Qt
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging


'''
The next two widgets (QTextEditLogger and LoggerTextArea) are used to create a textarea which also acts as handler for the logger objects created by the logging module
'''

class QTextEditLogger(logging.Handler):
#Code of this class was adapted from https://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt
    def __init__(self, textwidget, basic_format):
        super().__init__()
        formatter = Formatter(fmt=basic_format)
        self.setFormatter(formatter)
        self.textwidget = textwidget

    def emit(self, record):
        msg = self.format(record)
        self.textwidget.appendHtml(msg)

class Formatter(logging.Formatter):
    COLORS = {logging.INFO: QtGui.QColor("black"),
              logging.ERROR: QtGui.QColor("red"),
              logging.WARNING: QtGui.QColor("orange")}

    #def formatException(self, ei):
    #    result = super(Formatter, self).formatException(ei)
    #    return result

    def __init__(self, fmt="%(levelno)d: %(msg)s"):
        logging.Formatter.__init__(self, fmt=fmt)

    def format(self, record):
        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt
        # Replace the original format with one customized by logging level
        for level in self.COLORS.keys():
            if record.levelno == level:
                self._style._fmt = "<font color=\"{}\">{}</font>".format(self.COLORS[level].name(),format_orig)
        # Call the original formatter class to do the grunt work
        result = super().format(record)
        # Restore the original format configured by the user
        self._style._fmt = format_orig
        return result

        #s = super(Formatter, self).format(record)
        #if record.exc_text:
        #    s = s.replace('\n', '')
        #return s

class LoggerTextArea(Qt.QWidget):
#Code of this class was adapted from https://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt
    def __init__(self):
        super().__init__()

        self.widget = Qt.QPlainTextEdit()
        self.widget.setReadOnly(True)

        layout = Qt.QVBoxLayout()
        # Add the new logging box widget to the layout
        layout.addWidget(self.widget)
        self.setLayout(layout)
        self.widget.textChanged.connect(self.change_text)
        #self.ensureCursorVisible()

    def add_logger(self,logger):
        logTextBox = QTextEditLogger(self.widget, basic_format = logger.handlers[0].formatter._fmt)
        #if len(logger.handlers)>0:
        #    logTextBox.setFormatter(logger.handlers[0].formatter)
        logger.addHandler(logTextBox)

    def change_text(self):
        self.widget.verticalScrollBar().setValue(self.widget.verticalScrollBar().maximum())



import os
import PyQt5
dirname = os.path.dirname(PyQt5.__file__)
plugin_path = os.path.join(dirname, 'plugins', 'platforms')
os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = plugin_path
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import sys
import argparse
import importlib
import numpy as np
import inspect
from importlib.metadata import version  
import math

from ergastirio.experiment import experiment
from ergastirio.experiment_gui import experiment_gui
import ergastirio.widgets
import ergastirio.utils
import ergastirio.logger

## This is a quick and dirty solution to get rid of the warning "QWindowsWindow::setGeometry: Unable to set geometry ...", which is due to a bug of PyQt5 in Windows. 
def handler(msg_type, msg_log_context, msg_string):
    pass
PyQt5.QtCore.qInstallMessageHandler(handler)
##

graphics_dir = os.path.join(os.path.dirname(__file__), 'graphics')

class MainWindow(Qt.QMainWindow):
    '''
    The main window contains several menus and a QMdiArea object, 
    The QMdiArea object in turn contains subwindows where instruments, plots, data table and logs will be created
    Object creation:
        window = MainWindow()
    The QMdiArea object is saved as attribute:
        window.mdi 
    The mdi object is also the parent object which needs to be passed as parameter when creating an instance of the experiment object
    For example:
        Experiment = experiment(app,window,window.mdi,config_file)
    The mdi object contains a dictionary, window.mdi.containers whose elements represent the different subwindows.
    Specifically, window.mdi.containers = {'logging':{...},'tabledata':{...},'instruments':{...},'plots':{...} }
    The object
        window.mdi.containers[name]['container']
    is the widget that will act as container for the different part of the gui. 
    Each of this widget is a child (via setWidget() method) of a corresponding QScrollArea object, 
        window.mdi.containers[name]['scrollarea']
    In turns, the scrollarea objects are children (via setWidget() method) of corresponding QMdiSubWindow objects
        window.mdi.containers[name]['subwindow']
    '''
    _views =  {'View 1':{
                "instruments":  [ "11/20", "0", "9/20", "2/3" ],
                "plots":        [ "0", "0", "11/20", "2/3" ],
                "tabledata":    [ "0", "2/3", "2/3", "1/3" ],
                "logging":      [ "2/3", "2/3", "1/3", "1/3" ]
                }
               ,
               'View 2':{
                "instruments":  [ "2/3", "0", "1/3", "2/3" ],
                "plots":        [ "0", "0", "2/3", "2/3" ],
                "tabledata":    [ "0", "2/3", "2/3", "1/3" ],
                "logging":      [ "2/3", "2/3", "1/3", "1/3" ]
                }
               ,
               'View 3':{
                "instruments":  [ "1/2", "0", "1/2", "1/2"],
                "plots":        [ "0", "0", "1/2", "1/2" ],
                "tabledata":    [ "0", "1/2", "1/2", "1/2" ],
                "logging":      [ "1/2", "1/2", "1/2", "1/2" ]
                }
               ,
               'View 4':{
                "instruments":  [ "2/3", "0", "1/3", "1/2"],
                "plots":        [ "0", "0", "2/3", "1/2" ],
                "tabledata":    [ "0", "1/2", "2/3", "1/2" ],
                "logging":      [ "2/3", "1/2", "1/3", "1/2" ]
                }
               ,
               'View 5':{
                "instruments":  [ "2/3", "0", "1/3", "1/2"],
                "plots":        [ "0", "0", "2/3", "3/5" ],
                "tabledata":    [ "0", "3/5", "2/3", "2/5" ],
                "logging":      [ "2/3", "1/2", "1/3", "1/2" ]
                }
               }
    _default_view = 'View 4'
    #This dictionary will be used to create the subwindows of the mdi, and also to populate the "View" menu
    _mdi_basic_panels ={  'logging':{'title':'Logging'},
                    'tabledata':{'title':'Data acquisition'},
                    'instruments':{'title':'Instruments'},
                    'plots':{'title':'Plots'}
                      }
    _mdi_advanced_panels ={ 'connect_ramps':{'title':'Connect Ramps'}
                    }

    def __init__(self):
        super().__init__()
        self.setWindowTitle(f"{__package__} (v{version(__package__)})")
        self.mdi = Qt.QMdiArea()
        self.setCentralWidget(self.mdi)

        self.current_view = self._views[self._default_view]
       
        mdi_basic_panels = self._mdi_basic_panels
        mdi_advanced_panels = self._mdi_advanced_panels

        def initialize_panels(panels_dict):
        #Create the different subwindows, based on the element of the dictionary panels_dictionary
            for key,mdi_panel in panels_dict.items():
                panels_dict[key]['subwindow'] = Qt.QMdiSubWindow()
                panels_dict[key]['subwindow'].setWindowFlags(QtCore.Qt.CustomizeWindowHint)
                panels_dict[key]['subwindow'].setWindowTitle(panels_dict[key]['title'])
                panels_dict[key]['scrollarea'] = Qt.QScrollArea(self)
                panels_dict[key]['container'] = Qt.QWidget(self)
                panels_dict[key]['subwindow'].setWidget(panels_dict[key]['scrollarea'])
                panels_dict[key]['scrollarea'].setWidget(panels_dict[key]['container'])
                panels_dict[key]['scrollarea'].setWidgetResizable(True)

        initialize_panels(mdi_basic_panels)
        initialize_panels(mdi_advanced_panels)

        #Add subwindows to mdi object and make them visible
        for mdi_panel in mdi_basic_panels.values():
            self.mdi.addSubWindow(mdi_panel['subwindow'])
            mdi_panel['container'].show()

        #Add subwindows to mdi object, make it hidden and restore buttons
        for mdi_panel in mdi_advanced_panels.values():
            self.mdi.addSubWindow(mdi_panel['subwindow'])
            mdi_panel['subwindow'].setWindowFlags(QtCore.Qt.Window | QtCore.Qt.WindowMinimizeButtonHint)
            mdi_panel['subwindow'].setHidden(True)

        # Create top menus
        bar = self.menuBar()
        file = bar.addMenu("File")
        file.addAction("New Experiment")

        view = bar.addMenu("View")
        for k in self._views.keys():
            view.addAction(k)
        view.triggered[Qt.QAction].connect(self.action_view)

        showhide = bar.addMenu("Show/Hide")
        for mdi_panel in mdi_basic_panels.values():
            showhide.addAction(mdi_panel['title'])
        showhide.triggered[Qt.QAction].connect(self.action_showhide)

        advanced = bar.addMenu("Advanced")
        advanced.addAction("Connect Ramps")
        advanced.triggered[Qt.QAction].connect(self.action_showhide)
        
        # Store mdi_basic_panels and mdi_advanced_panels as attributes of the window object
        self.basic_panels = mdi_basic_panels
        self.advanced_panels = mdi_advanced_panels
        self.all_panels = {**(self.basic_panels),**(self.advanced_panels)}
        self.set_view(self.current_view)

        # Create Logging area
        box = Qt.QVBoxLayout()
        self.logging_text_area = ergastirio.logger.LoggerTextArea()
        box.addWidget(self.logging_text_area)
        self.basic_panels['logging']['container'].setLayout(box)

    def action_view(self, p):
        for k in self._views.keys():
            if p.text() == k:
                self.current_view = self._views[k]
                self.set_view(self._views[k])
        return

    def action_showhide(self,p):
        for mdi_panel in self.all_panels.values():
            if p.text() == mdi_panel['title']:
                mdi_panel['subwindow'].setHidden(not mdi_panel['subwindow'].isHidden())

    def resizeEvent(self, event):
        Qt.QMainWindow.resizeEvent(self, event)
        self.set_view(self.current_view)

    def close(self, experiment):
        current_view = self.get_current_view()
        experiment.settings['GUI']['View'] = current_view
        experiment.close()
        pass

    def get_current_view(self):
        geomMDI = self.mdi.geometry()
        MDI_width = geomMDI.width()
        MDI_height = geomMDI.height()
        MDI_x = geomMDI.x()
        MDI_y = geomMDI.y()
        view = dict()
        for key in self.all_panels.keys():
            view[key] = list(self.all_panels[key]['subwindow'].geometry().getRect())
            view[key][0] = view[key][0]/MDI_width
            view[key][2] = view[key][2]/MDI_width
            view[key][1] = view[key][1]/MDI_height
            view[key][3] = view[key][3]/MDI_height
        return view

    def set_view(self,view):
        geomMDI = self.mdi.geometry()
        MDI_width = geomMDI.width()
        MDI_height = geomMDI.height()
        MDI_x = geomMDI.x()
        MDI_y = geomMDI.y()
        for panel,geom in view.items():
            geom_abs = [] 
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[0])*MDI_width)  )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[1])*MDI_height) )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[2])*MDI_width)  )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[3])*MDI_height) )
            self.all_panels[panel]['subwindow'].setGeometry(*geom_abs)
        self.current_view = self.get_current_view()


def main():
    parser = argparse.ArgumentParser(description = "",epilog = "")
    parser.add_argument('-e', 
                        help=f"Path of .json file contaning the configuration of this experiment",
                        action="store", dest="config", type=str, default=None)
    parser.add_argument("-s", "--decrease_verbose", help="Decrease verbosity.", action="store_true")
    parser.add_argument('-virtual', help=f"Use virtual driver for each instrument (when available).", action="store_true")
    args = parser.parse_args()
    virtual = args.virtual
    if args.config:
        config_file = os.path.abspath(args.config)
    else:
        folder_default_file = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(folder_default_file,'config_default.json')
    
    app = Qt.QApplication(sys.argv)
    window = MainWindow()
    window.show()

    spinner = ergastirio.widgets.QtWaitingSpinner(window)
    spinner.setInnerRadius(30)
    spinner.setLineLength(30)
    spinner.setLineWidth(10)
    spinner.setRevolutionsPerSecond(1)
    spinner.setColor(QtGui.QColor(255, 255, 255))
    spinner.start() # starts spinning

    #app.aboutToQuit.connect(window.closeEvent) 

    #Create a new experiment
    Experiment = experiment(app)
    success = Experiment.initiate(config_file=config_file, func_logger_handler_adder = window.logging_text_area.add_logger, virtual=virtual)
                                                # func_logger_handler_adder must be defined as a function that takes one argument as input. The input argument is a Logger object
                                                # func_logger_handler_adder will add any Logger passed to it an handler defined elsewhere
    window.set_view(Experiment.settings['GUI']['View'])  

    #Generate the GUI for the new experiment
    if success:
        Experiment_gui = experiment_gui(Experiment,parent=window)

    spinner.stop()
    app.aboutToQuit.connect(lambda:window.close(Experiment)) 
    
    app.exec()# Start the event loop.

if __name__ == '__main__':
    main()


'''
This module contains several function which are used to initialize a certain experiment. 
Before using the functions in this module, an experiment() object needs to be created (typically by the main.pyw file), and a config file needs to be assigned to 
the experiment. After a config file have been assigned to the experiment, the function load_config_and_setup_exp(exp) is called, where exp is the experiment() object.
'''
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import numpy as np
import importlib
#from importlib.metadata import version  
import json
import ergastirio.utils
 
def set_up_experiment(exp):
    '''
    This function initializes a certain experiment (passed as an input argument) based on the settings stored in the dictionary exp.settings.
    It subsequently calls several functions to load instruments, initialize storage arrays, validate user configs
    '''
        
    if not load_instruments(exp):       #Load all instruments based on the strings specified in the list exp.settings['Instruments_Settings']
        return False

    #if not load_all_ramps(exp):         #Check which instrument has a ramp object
    #    return False

    if not generate_data_headers(exp):  #Based on the loaded instruments, and on the data created by each instrument, it populates the list of strings exp.data_headers
        return False

    if not load_and_check_postprocessed_data(exp):
        return False

    generate_list_valid_data_names(exp) # Use headers, plus "acq#", to generate a list of names that can be used for plotting and post-processing

    initialize_storage_variable(exp)    #Create Storage variables

    initialize_acquisition_settings(exp)
       
    if not validate_plots_settings(exp): #check if the plots settings specified in the .json file are valid (i.e. any quantity defined in 'x' and 'y' is a valid device data
        return False
    return True

def load_instruments(exp):
    ''' 
    Loads all instruments defined in exp.settings['Instruments_Settings']['instruments'] and stores them in the list exp.instruments
    For each intruments, the respective interface object is also created, but not the gui.

    Returns true if all instruments were loaded correctly, false otherwise
    '''
    exp.logger.info(f"Loading all instruments specified in {exp.config_file}...")
    exp.instruments = []
    i=0
    for instrument in exp.settings['Instruments_Settings']['instruments'] :

        # Check if the name of instrument corresponds to an install package ## TO DO : check also that it is package supported by ergastirio
        if (importlib.util.find_spec(instrument) is None):
            exp.logger.error(f"{instrument} is not a valid python package. Fix this error and restart this application")
            return False
        device_module =  importlib.import_module(instrument+'.main')

        # Check that the main module of this instrument contains a class interface
        if not(hasattr(device_module,'interface')):
            exp.logger.error(f"{instrument} is a valid package, but it does not contain an 'interface' class.")
            return False

        fullname = f"dev{i}_{instrument}"

        try:
            exp.logger.info(f"Loading an interface for instrument {instrument}, with name : {fullname}")
            config_dict=None
            if len(exp.settings["Instruments_Settings"]["settings_each_instrument"]) > i:
                config_dict=exp.settings["Instruments_Settings"]["settings_each_instrument"][i]
                if isinstance(exp.settings["Instruments_Settings"]["settings_each_instrument"][i],dict):
                    exp.logger.info(f"Found settings for this interface in the file {exp.config_file}.")
                else:
                    config_dict = None    
            exp.instruments.append(
                                        {    
                                        'name':f"dev{i}",
                                        'fullname':fullname,
                                        'type':instrument,
                                        'device_module' : device_module,
                                        'interface_class':device_module.interface,
                                        'interface':  device_module.interface(  app=exp.app,
                                                                                name_logger=fullname,
                                                                                config_dict=config_dict,
                                                                                virtual=exp.virtual_modality) 
                                        }
                                    )
            exp.func_logger_handler_adder(logger=logging.getLogger(exp.instruments[-1]['fullname'])) #Having this here is kind of ugly because it partially breaks the model-view philosophy; on the other hand it allows catching the logs immediately
        except Exception as e:
            exp.logger.error(f"An error occurred while loading the instrument {instrument}: {e}")
            return False
        
        i = i+1
    return True

def generate_data_headers(exp):
    ''' 
    Generates a list of data headers for the experiment exp, based on the loaded instruments.
    For each instrument in exp.instruments, the list of output data is contained in the keys of the dictionary instrument.['interface_class'].output.
    The headers 'timestamp' and 'time' are also added at the beginning of the data headers.

    Returns True
    '''
    exp.logger.info(f"Preparing the data headers...")
    exp.data_headers =[]
    exp.data_headers.append('timestamp')
    exp.data_headers.append('time')
    for instrument in exp.instruments:
        if hasattr(instrument['interface_class'],'output'):
            for key in instrument['interface_class'].output.keys():
                exp.data_headers.append(instrument['name']+'.'+key)
        else:
            exp.logger.info(f"Instrument {instrument['fullname']} does not have an \'output\' dictionary defined. I wil assume that it does not produce data.")
    exp.logger.info(f"The following data will be acquired: {exp.data_headers}")

    return True

def load_and_check_postprocessed_data(exp):
    ''' 
    Load the post-processed data defined in exp.settings['Instruments_Settings']['post_processed_data'] (if any is defined)
    and checks that each of them represents a valid formula.
    It appends the name of each post-processed data to the list exp.data_headers.

    Returns true if all post-processed data are valid, false otherwhise.

    TO DO: check that no post-processed data has same name as another valid data.
    '''
    if ('post_processed_data' in exp.settings['Instruments_Settings'].keys()):
        exp.logger.info(f"Validating post-processed data...")
        for post_processed_data_name,post_processed_data in exp.settings['Instruments_Settings']['post_processed_data'].items():
            (msg,flag) = ergastirio.utils.check_list_is_valid_formula(post_processed_data,valid_variable_names=exp.data_headers)
            if not(flag==1):
                exp.logger.error(f"A problem occurred with the post-processed data: {msg}.")
                return False
            exp.data_headers.append(post_processed_data_name)

    return True

def generate_list_valid_data_names(exp):
    exp.valid_data_names = exp.data_headers.copy()    #The list of headers gives a list of valid data can be plotted. This assume that, for each row of data, each entry is a scalar
    exp.valid_data_names.append("acq#")               #Need to implement this better
    
def initialize_storage_variable(exp):
    ''' Initialize all arrays for data storage. '''      
    exp.data = ergastirio.classes.EnhancedList([]) #np.empty([0,len(exp.data_headers)]) 
    exp.data_std = ergastirio.classes.EnhancedList([])
    return True

def initialize_acquisition_settings(exp):
    if 'trigger_modality' in exp.settings['General_Settings'].keys():
        exp.trigger_modality = exp.settings['General_Settings']['trigger_modality']
    else:
        return
    if 'trigger_instrument' in exp.settings['Instruments_Settings'].keys():
        if 'trigger_delay' in exp.settings['Instruments_Settings'].keys():
            delay = exp.settings['Instruments_Settings']['trigger_delay']
        else:
            delay = 0
        exp.set_trigger_instrument(exp.settings['Instruments_Settings']['trigger_instrument'], delay = delay)

    
def validate_plots_settings(exp):
    '''
    Looks at the content of exp.settings['Plot_Settings']['plots'] and check that all settings specified by the user are valid. If the user specificied an empty plot,
    it uses "acq#" for x axis and the first valid data (excluding 'timestamp' and 'time') for the y axis
    '''
    exp.logger.info(f"Validating all plot settings...")
    exp.plots = []

    for plotindex, plot in enumerate(exp.settings['Plot_Settings']['plots']):
        if isinstance(plot,dict):
            if 'x' in plot.keys():
                if not plot['x'] in exp.valid_data_names:
                    exp.logger.error(f"Plot #{plotindex}: {plot['x']} is not a valid device data")
                    return False
            else:
                plot['x'] =  exp.valid_data_names[-1]
            if 'y' in plot.keys():
                if isinstance(plot['y'],str):
                    plot['y'] = [plot['y']]
                for y_data in plot['y']:
                    if not y_data in exp.valid_data_names:
                        exp.logger.error(f"Plot #{plotindex}: {y_data} is not a valid device data")
                        return False    
            else:
                plot['y'] =  [exp.valid_data_names[2]]
        else:
            exp.logger.error(f"Each plot must be defined in the .json file as a dictionary containing an 'x' and a 'y' key.")
            return False
        exp.plots.append({'name':f"plot{plotindex}",
                            'x':plot['x'],
                            'y':plot['y']})
        exp.logger.info(f"Plot #{plotindex}, x = {plot['x']}, y = {plot['y']}")
    return True




'''
This module contains several customized widgets used by Ergastirio
'''
import PyQt5.QtWidgets as Qt
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import numpy as np
import pyqtgraph as pg
import os

graphics_dir = os.path.join(os.path.dirname(__file__), '../graphics')

class PlotObject:
    ''' Interactive plot. It has built-in properties 'data' and 'data_headers', and it updates automatically the content of the plot everytime
        this properties are updated.
    '''
    _data_headers =[]
    _data = []
    _data_error = []
    _colors = 'bgrcmyw'
    _styles = [pg.mkPen(c, width=2, style=QtCore.Qt.SolidLine) for c in _colors]
    _symbol = 'o'
    _style_labels = {"color": "#fff", "font-size": "20px"}
    def __init__(self, parent, data_headers, plot_config):
        '''
        parent        = a QWidget (or QMainWindow) object that will be the parent for this plot
        data_headers
        plot_config   = a dictionary specifying the settings of this plot. The dictionary must contain at least two keys, 'x' and 'y'. 
                        The value of 'x' must be a single string. The value of 'y' can be either a string or a list of strings.
        '''

        self.parent = parent
        self._data_headers =[]
        self._data = []
        self._data_error = []
        self.data_headers = data_headers #Note: here we assume that data_headers do not already contain the "acq#" field. It will be added dynamically by counting the number of rows
        self.plot_config = plot_config

        self.Max = 0 #Keep track of the maximum of minimum values plotted in this plot (among all possible curves). It is used for resizing purposes
        self.Min = 0

        self.graphWidget = self.create_plot()
        self.controlWidget = self.create_controlPanel()

        vbox = Qt.QVBoxLayout()
        vbox.addWidget(self.graphWidget) 
        vbox.addWidget(self.controlWidget) 
        vbox.setSpacing(0)
        #vbox.addStretch(1)
        self.parent.setLayout(vbox)
        self.set_quantities_to_plot() #This will initialize axis and legend of the plot based on the current value of self.plot_config
        self.refresh_plot()
    
    @property
    def data_headers(self):
        return self._data_headers
    @data_headers.setter
    def data_headers(self,h):
        self._data_headers = h
        self._data_headers_full = h.copy()
        self._data_headers_full.insert(0,"acq#")
        self._data_headers_full_radiobuttons_x_axis =[] # these two lists will store the radiobuttons and checkbox used to decide which quantity is plotted 
        self._data_headers_full_checkboxs_y_axis =[]    # on x and y axis. Each element of self._data_headers_full is associated to one radiobutton and one checkbox

    @property
    def data(self):
        return self._data
    @data.setter
    def data(self,d):
        # Every time the list data is updated, store it and update the plots
        self._data = d
        self.refresh_plot()

    @property
    def data_error(self):
        return self._data_error
    @data_error.setter
    def data_error(self,d):
        self._data_error = d
        self.refresh_plot()

    ### GUI
    def create_plot(self):
        graphWidget = pg.PlotWidget()
        graphWidget.showGrid(x=True, y=True)
        #graphWidget.setMenuEnabled(False)
        return graphWidget

    def create_controlPanel(self):
        w = Qt.QWidget()
        hbox = Qt.QHBoxLayout()
        w.button_home = Qt.QPushButton("")
        w.button_home.setIcon(QtGui.QIcon(os.path.join(graphics_dir,'home.png')))
        w.button_home.setFixedSize( 30, 30)

        w.toolButton_X = Qt.QToolButton()
        w.toolButton_X.setText('X')
        w.toolButton_X.show()
        #w.toolButton_X.setFixedSize(w.button_home.size())
        #w.toolButton_X.setSizePolicy(Qt.QSizePolicy.Preferred,Qt.QSizePolicy.Expanding)
        w.toolButton_X.setFixedSize( 30, 30)

        w.toolButton_Y = Qt.QToolButton()
        w.toolButton_Y.setText('Y')
        w.toolButton_Y.show()
        #w.toolButton_Y.setSizePolicy(Qt.QSizePolicy.Preferred,Qt.QSizePolicy.Expanding)
        w.toolButton_Y.setFixedSize( 30, 30)

        ### Connect widgets to functions
        self.create_menu_XY(w.toolButton_X,w.toolButton_Y)
        w.button_home.clicked.connect(self.reset_plot_view)
        ###

        hbox.addWidget(w.button_home)
        hbox.addWidget(w.toolButton_X)
        hbox.addWidget(w.toolButton_Y)

        hbox.addStretch(1)
        hbox.setSpacing(0)
        hbox.setContentsMargins(0, 0, 0, 0)
        w.setLayout(hbox)
        return w

    def create_menu_XY(self,toolButton_X,toolButton_Y):
        #code inspired by an answer to https://stackoverflow.com/questions/34731826/keep-menu-open-after-clicking-on-the-button-it-is-launched-with
        toolMenuX = Qt.QMenu()
        toolMenuY = Qt.QMenu()
        self._data_headers_full_radiobuttons_x_axis = []
        self._data_headers_full_checkboxs_y_axis = []
        #self.radio_x = []
        for data_name in self._data_headers_full:
            radiobutton = Qt.QRadioButton(data_name,toolMenuX)
            radiobutton.value = data_name
            if data_name == self.plot_config['x']:
                radiobutton.setChecked(True)
            radiobutton.toggled.connect(lambda x,object=radiobutton: self.update_plotted_data('x',object))
            checkableAction = Qt.QWidgetAction(toolMenuX)
            checkableAction.setDefaultWidget(radiobutton)
            toolMenuX.addAction(checkableAction)
            self._data_headers_full_radiobuttons_x_axis.append(radiobutton)

            checkbox = Qt.QCheckBox(data_name,toolMenuY)
            checkbox.value = data_name
            if data_name in self.plot_config['y']:
                checkbox.setChecked(True)
            checkbox.stateChanged.connect(lambda: self.update_plotted_data('y'))
            checkableAction = Qt.QWidgetAction(toolMenuY)
            checkableAction.setDefaultWidget(checkbox)
            toolMenuY.addAction(checkableAction)
            self._data_headers_full_checkboxs_y_axis.append(checkbox)

        toolButton_X.toolMenu = toolButton_X
        toolButton_Y.toolMenu = toolButton_Y
        toolButton_X.setMenu(toolMenuX)
        toolButton_Y.setMenu(toolMenuY)
        toolButton_X.setPopupMode(Qt.QToolButton.InstantPopup)
        toolButton_X.setStyleSheet('QToolButton::menu-indicator { image: none; }')
        toolButton_Y.setPopupMode(Qt.QToolButton.InstantPopup)
        toolButton_Y.setStyleSheet('QToolButton::menu-indicator { image: none; }')
    ### END GUI

    ### GUI Events Functions
    def update_plotted_data(self, axis, object = None):
        '''
        Based on the current values of the radiobuttons and checkbox in the toolbutton menu, it updates 
        the content of self.plot_config['x'] and self.plot_config['y']
        '''
        if axis == 'x':
            if object.isChecked() == True:
                self.plot_config['x'] = object.value #(need to implement this better)
        if axis == 'y':
            self.plot_config['y'] = []
            for checkbox in self._data_headers_full_checkboxs_y_axis:
                if checkbox.isChecked() == True:
                    self.plot_config['y'].append(checkbox.value)
        self.set_quantities_to_plot()
        self.refresh_plot()

    def reset_plot_view(self):
        self.graphWidget.autoRange()
        self.graphWidget.enableAutoRange()

    ### END GUI Events Functions

    def set_quantities_to_plot(self):
        '''
        Based on the content of self.plot_config, it updates the values of self.x_index and self.y_index .
        self.x_index corresponds to the column index of the self._data matrix which will be used as x-quantity
        self.y_index is a list of indices, corresponding to the columns of the self._data matrix which will be used as y-quantities
        '''
        self.x_name = self.plot_config['x'] 
        self.x_index = self._data_headers_full.index(self.x_name) - 1 #in the self._data_headers_full list, we have added the "acq#" element at the beginning
                                                                      #thus, we substract 1 to the index to match this index with the column index in self._data
        self.y_index =[]
        self.y_names = self.plot_config['y']
        for y_name in self.plot_config['y']:
            self.y_index.append(self._data_headers_full.index(y_name) - 1) #in the self._data_headers_full list, we have added the "acq#" element at the beginning
        return

    def refresh_plot(self):
        self.graphWidget.clear()
        if len(self.y_index) == 0: #if len(self.y_index) == 0 then there are no quantities to plot on y axis. Thus, we don't do anything in this routine
            return
        self.graphWidget.addLegend()

        if self.x_index == -1:
            x = list(range(1,len(self._data)+1)) #if self.x_index=-1, then the data to use on x axis is the acquisition number. We build the x-data explicitly.
        else:
            x = [row[self.x_index] for row in self._data] #otherwise,  the data to use on x axis is one of the columns of self._data. We extract it.
        
        self.graphWidget.setLabel('bottom', self.x_name ,**self._style_labels)

        for i,y_index in enumerate(self.y_index):
            if y_index == -1:
                y = list(range(1,len(self._data)+1)) #if self.y_index=-1, then this specific y-data is the acquisition number. We build the ydata explicitly.
            else:
                y = [row[y_index] for row in self._data] #otherwise, this specific y-data is one of the columns of self._data. We extract it.
            index_style = (y_index+1) % len(self._styles)
            try:
                self.graphWidget.plot(x,y,  name=self.y_names[i],    pen=self._styles[index_style],    symbol=self._symbol,    symbolBrush=self._colors[index_style])
            except Exception as e:
                print(f"An error occurred while trying to plot\n: {e}")
                return
            
 



class EnhancedList(list):
    '''
    This class is used to generate an "event-based" list object. Everytime the content of the list changes, it also stores the list 
    in a certain property of all the objects specified in the list linked_objects. 
    Each element of linked_objects is a two-element list in the form [class_instance,class_property]. This behavior is useful when the linked objects
    are, e.g., plots and tables, and the targeted property is defined via a @setter, in order to automatically update parts of the gui

    Examples:

        class test():
            def __init__( self ):
                self.__value = "old value"

            @property
            def value( self ):
                return self.__value

            @value.setter
            def value( self, value ):
                self.__value = value
                print("Targeted list changed to " + str(value))

        a = EnhancedList([1,2,3])
        t=test()
        a.add_syncronized_objects([t,test.value])
        a.append(4)

    Targeted list changed to [1, 2, 3, 4]

    '''

    def __init__(self,  *args):
        self.linked_objects = []

    def add_syncronized_objects(self,list_objects):
        self.linked_objects.append(list_objects)

    def sync(self):
        for obj in self.linked_objects:
            obj[1].fset(obj[0], self)

    def sync_copy(self):
        for obj in self.linked_objects:
            obj[1].fset(obj[0], self.copy())

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.sync()
    def __delitem__(self, value):
        super().__delitem__(value)
        self.sync()
    def __add__(self, value):
        super().__add__(value)
        self.sync()
    def __iadd__(self, value):
        super().__iadd__(value)
        self.sync()
    def append(self, value):
        super().append(value)
        self.sync()
    def remove(self, value):
        super().remove(value)
        self.sync()
    def insert(self, *args):
        super().insert(*args)
        self.sync()
    def pop(self, *args):
        super().pop(*args)
        self.sync()
    def clear(self):
        super().clear()
        self.sync()